//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "parallel/PetscHelper.h"
#include "parallel/PetscSolver.h"
#include "Global.h"

namespace AMDiS {

  namespace petsc_helper {

    using namespace std;
    
    void getMatLocalColumn(Mat mat, PetscMatCol &matCol)
    {
      PetscInt firstIndex, lastIndex;
      MatGetOwnershipRange(mat, &firstIndex, &lastIndex);
    
      PetscInt nCols;
      const PetscInt *cols;
      const PetscScalar *values;
      
      for (int row = firstIndex; row < lastIndex; row++) {
	MatGetRow(mat, row, &nCols, &cols, &values);
	
	for (int i = 0; i < nCols; i++) {
	  if (values[i] != 0.0) {
	    matCol[cols[i]].first.push_back(row - firstIndex);
	    matCol[cols[i]].second.push_back(values[i]);
	  }
	}

	MatRestoreRow(mat, row, &nCols, &cols, &values);
      }
    }


    void setMatLocalColumn(Mat mat, int column, Vec vec)
    {
      PetscInt firstIndex;
      MatGetOwnershipRange(mat, &firstIndex, PETSC_NULL);

      PetscInt vecSize;
      VecGetLocalSize(vec, &vecSize);

      PetscScalar *tmpValues;
      VecGetArray(vec, &tmpValues);
      for (int i  = 0; i < vecSize; i++)
	MatSetValue(mat, 
		    firstIndex + i,
		    column,
		    tmpValues[i],
		    ADD_VALUES);
      VecRestoreArray(vec, &tmpValues);
    }


    void getColumnVec(const SparseCol &matCol, Vec vec)
    {
      VecSet(vec, 0.0);
      VecSetValues(vec, matCol.first.size(), 
		   &(matCol.first[0]), &(matCol.second[0]), INSERT_VALUES);
      VecAssemblyBegin(vec);
      VecAssemblyEnd(vec);
    }


    void blockMatMatSolve(KSP ksp, Mat mat0, Mat &mat1)
    {
      // === We have to  calculate mat1 = ksp mat0:                       ===
      // ===    - get all local column vectors from mat0                  ===
      // ===    - solve with ksp for this column vector as the rhs vector ===
      // ===    - set the result to the corresponding column vector of    ===
      // ===      matrix mat1                                             ===
      
      PetscInt localRows, localCols, globalRows, globalCols;
      MatGetLocalSize(mat0, &localRows, &localCols);
      MatGetSize(mat0, &globalRows, &globalCols);

      MatCreateAIJ(PETSC_COMM_WORLD,
		   localRows, localCols, globalRows, globalCols,
		   150, PETSC_NULL, 150, PETSC_NULL, &mat1);
      MatSetOption(mat1, MAT_NEW_NONZERO_ALLOCATION_ERR, PETSC_FALSE);
      
      // Transform matrix mat0 into a sparse column format.
      PetscMatCol mat0_cols;
      getMatLocalColumn(mat0, mat0_cols);
      
      Vec tmpVec;
      VecCreateSeq(PETSC_COMM_SELF, localRows, &tmpVec);
      
      // Solve for all column vectors of mat A_BPi and create matrix matK
      for (PetscMatCol::iterator it = mat0_cols.begin(); 
	   it != mat0_cols.end(); ++it) {
	getColumnVec(it->second, tmpVec);
	KSPSolve(ksp, tmpVec, tmpVec);
	setMatLocalColumn(mat1, it->first, tmpVec);
      }
      
      VecDestroy(&tmpVec);
      
      MatAssemblyBegin(mat1, MAT_FINAL_ASSEMBLY);
      MatAssemblyEnd(mat1, MAT_FINAL_ASSEMBLY);
    }


    void matNestConvert(Mat matNest, Mat &mat)  
    {
      FUNCNAME("matNestConvert()");

      PetscInt nestRows, nestCols;
      MatNestGetSize(matNest, &nestRows, &nestCols);

      TEST_EXIT(nestRows == 2 && nestCols == 2)
	("This funciton is only implemented for 2x2 nested matrices!\n");

      Mat mat00, mat01, mat10, mat11;
      MatNestGetSubMat(matNest, 0, 0, &mat00);
      MatNestGetSubMat(matNest, 0, 1, &mat01);
      MatNestGetSubMat(matNest, 1, 0, &mat10);
      MatNestGetSubMat(matNest, 1, 1, &mat11);

      int nRankRows0, nOverallRows0;
      MatGetLocalSize(mat00, &nRankRows0, PETSC_NULL);
      MatGetSize(mat00, &nOverallRows0, PETSC_NULL);

      int nRankRows1, nOverallRows1;
      MatGetLocalSize(mat11, &nRankRows1, PETSC_NULL);
      MatGetSize(mat11, &nOverallRows1, PETSC_NULL);

      int firstRow0;
      MatGetOwnershipRange(mat00, &firstRow0, PETSC_NULL);

      int firstRow1;
      MatGetOwnershipRange(mat11, &firstRow1, PETSC_NULL);

      int nRankRows = nRankRows0 + nRankRows1;
      int nOverallRows = nOverallRows0 + nOverallRows1;
      int firstRow = firstRow0 + firstRow1;

      int mpiSize = MPI::COMM_WORLD.Get_size();
      int mpiRank = MPI::COMM_WORLD.Get_rank();
      vector<int> allFirstRow0(mpiSize + 1, 0);
      vector<int> allFirstRow1(mpiSize + 1, 0);
      MPI::COMM_WORLD.Allgather(&nRankRows0, 1, MPI_INT, &(allFirstRow0[1]), 1, MPI_INT);
      MPI::COMM_WORLD.Allgather(&nRankRows1, 1, MPI_INT, &(allFirstRow1[1]), 1, MPI_INT);

      for (int i = 1; i < mpiSize + 1; i++) {
	allFirstRow0[i] += allFirstRow0[i - 1];
	allFirstRow1[i] += allFirstRow1[i - 1];
      }

      TEST_EXIT_DBG(allFirstRow0[mpiRank] == firstRow0)("Should not happen!\n");
      TEST_EXIT_DBG(allFirstRow1[mpiRank] == firstRow1)("Should not happen!\n");

      MatCreateAIJ(PETSC_COMM_WORLD, 
		   nRankRows, nRankRows,
		   nOverallRows, nOverallRows,
		   25, PETSC_NULL, 25, PETSC_NULL, &mat);
      MatSetOption(mat, MAT_NEW_NONZERO_ALLOCATION_ERR, PETSC_FALSE);

      PetscInt nCols;
      const PetscInt *cols;
      const PetscScalar *vals;

      for (int i = 0; i < nRankRows0; i++) {
	PetscInt newRowIndex = firstRow + i;

	MatGetRow(mat00, firstRow0 + i, &nCols, &cols, &vals);
	for (int j = 0; j < nCols; j++) {
	  int rank = -1;
	  for (int k = 0; k < mpiSize; k++) {
	    if (cols[j] >= allFirstRow0[k] && cols[j] < allFirstRow0[k + 1]) {
	      rank = k;
	      break;
	    }
	  }
	  TEST_EXIT_DBG(rank != -1)("Should not happen!\n");

	  PetscInt newColIndex = cols[j] + allFirstRow1[rank];
	  MatSetValue(mat, newRowIndex, newColIndex, vals[j], INSERT_VALUES);
	}
	MatRestoreRow(mat00, firstRow0 + i, &nCols, &cols, &vals);

	MatGetRow(mat01, firstRow0 + i, &nCols, &cols, &vals);
	for (int j = 0; j < nCols; j++) {
	  int rank = -1;
	  for (int k = 0; k < mpiSize; k++) {
	    if (cols[j] >= allFirstRow1[k] && cols[j] < allFirstRow1[k + 1]) {
	      rank = k;
	      break;
	    }
	  }
	  TEST_EXIT_DBG(rank != -1)("Should not happen!\n");

	  PetscInt newColIndex = cols[j] + allFirstRow0[rank + 1];
	  MatSetValue(mat, newRowIndex, newColIndex, vals[j], INSERT_VALUES);
	}
	MatRestoreRow(mat01, firstRow0 + i, &nCols, &cols, &vals);	
      }

      for (int i = 0; i < nRankRows1; i++) {
	PetscInt newRowIndex = firstRow + nRankRows0 + i;

	MatGetRow(mat10, firstRow1 + i, &nCols, &cols, &vals);
	for (int j = 0; j < nCols; j++) {
	  int rank = -1;
	  for (int k = 0; k < mpiSize; k++) {
	    if (cols[j] >= allFirstRow0[k] && cols[j] < allFirstRow0[k + 1]) {
	      rank = k;
	      break;
	    }
	  }
	  TEST_EXIT_DBG(rank != -1)("Should not happen!\n");

	  PetscInt newColIndex = cols[j] + allFirstRow1[rank];
	  MatSetValue(mat, newRowIndex, newColIndex, vals[j], INSERT_VALUES);
	}
	MatRestoreRow(mat10, firstRow1 + i, &nCols, &cols, &vals);

	MatGetRow(mat11, firstRow1 + i, &nCols, &cols, &vals);
	for (int j = 0; j < nCols; j++) {
	  int rank = -1;
	  for (int k = 0; k < mpiSize; k++) {
	    if (cols[j] >= allFirstRow1[k] && cols[j] < allFirstRow1[k + 1]) {
	      rank = k;
	      break;
	    }
	  }
	  TEST_EXIT_DBG(rank != -1)("Should not happen!\n");

	  PetscInt newColIndex = cols[j] + allFirstRow0[rank + 1];
	  MatSetValue(mat, newRowIndex, newColIndex, vals[j], INSERT_VALUES);
	}
	MatRestoreRow(mat11, firstRow1 + i, &nCols, &cols, &vals);
      }

      MatAssemblyBegin(mat, MAT_FINAL_ASSEMBLY);
      MatAssemblyEnd(mat, MAT_FINAL_ASSEMBLY);
    }


    void setSolver(KSP ksp, 
		   string kspPrefix,
		   KSPType kspType, 
		   PCType pcType, 
		   PetscReal rtol,
		   PetscReal atol,
		   PetscInt maxIt)
    {
      KSPSetType(ksp, kspType);
      KSPSetTolerances(ksp, rtol, atol, PETSC_DEFAULT, maxIt);
      if (kspPrefix != "")
	KSPSetOptionsPrefix(ksp, kspPrefix.c_str());
      KSPSetFromOptions(ksp);

      PC pc;
      KSPGetPC(ksp, &pc);
      PCSetType(pc, pcType);
    }
  }

}
